""" Schemas used by the sample app"""
from protean.core.repository import repo
from protean.impl.repository.dict_repo import DictSchema

from .entities import Dog
from .entities import Human


class DogSchema(DictSchema):
    """ Schema for the Dog Entity"""

    class Meta:
        """ Meta class for schema options"""
        entity = Dog
        schema_name = 'dogs'


class HumanSchema(DictSchema):
    """ Schema for the Human Entity Entity"""

    class Meta:
        """ Meta class for schema options"""
        entity = Human
        schema_name = 'humans'


repo.register(DogSchema)
repo.register(HumanSchema)



""" Sample Protean Flask app for testing"""
from flask import Flask

from protean_flask import Protean

from .blueprint import blueprint
from .views import CreateDogResource
from .views import CurrentContextResource
from .views import DeleteDogResource
from .views import HumanResourceSet
from .views import ListDogResource
from .views import ShowDogResource
from .views import UpdateDogResource
from .views import flask_view

app = Flask(__name__)
api = Protean(app)

app.add_url_rule('/dogs/<int:identifier>',
                 view_func=ShowDogResource.as_view('show_dog'),
                 methods=['GET'])
app.add_url_rule('/dogs',
                 view_func=CreateDogResource.as_view('create_dog'),
                 methods=['POST'])
app.add_url_rule('/dogs',
                 view_func=ListDogResource.as_view('list_dogs'),
                 methods=['GET'])
app.add_url_rule('/dogs/<int:identifier>',
                 view_func=UpdateDogResource.as_view('update_dog'),
                 methods=['PUT'])
app.add_url_rule('/dogs/<int:identifier>',
                 view_func=DeleteDogResource.as_view('delete_dog'),
                 methods=['DELETE'])
app.add_url_rule('/flask-view', view_func=flask_view,
                 methods=['GET'])
app.add_url_rule('/current-context', methods=['GET'],
                 view_func=CurrentContextResource.as_view('current_context'))\

api.register_viewset(HumanResourceSet, 'humans', '/humans', pk_type='int',
                     additional_routes=['/<int:identifier>/my_dogs'])

app.register_blueprint(blueprint, url_prefix='/blueprint')



""" Serializers used by the sample app """
from protean_flask.core.serializers import EntitySerializer

from .entities import Dog
from .entities import Human


class DogSerializer(EntitySerializer):
    """ Serializer for Dog Entity"""

    class Meta:
        entity = Dog


class HumanSerializer(EntitySerializer):
    """ Serializer for Human Entity"""

    class Meta:
        entity = Human



""" Module for defining the response Renderers """
from flask import jsonify
from flask import make_response


def render_json(data, code, headers):
    """ Render the response as a JSON """

    resp = make_response(jsonify(data), code)
    resp.headers.extend(headers or {})

    return resp



"""Module to setup Factories and other required artifacts for tests"""

import os

import mock
import pytest

os.environ['PROTEAN_CONFIG'] = 'tests.support.sample_config'


@pytest.fixture(scope='module', autouse=True)
@mock.patch.dict(os.environ, {'PROTEAN_CONFIG': 'tests.support.sample_config'})
def config():
    """Global Config fixture for all tests"""

    from protean.conf import active_config
    return active_config



"""Module to test Serializer functionality"""

import marshmallow as ma
import pytest
from protean.core.exceptions import ConfigurationError

from protean_flask.core.serializers import EntitySerializer

from ..support.sample_app.entities import Dog


class DogSerializer(EntitySerializer):
    """ Serializer for the Dog Entity """
    class Meta:
        entity = Dog


class TestEntitySerializer:
    """Tests for EntitySerializer class"""

    def test_init(self):
        """Test initialization of EntitySerializer derived class"""
        s = DogSerializer()
        assert s is not None

        # Check that the entity gets serialized correctly
        s_result = s.dump(Dog(id=1, name='Johnny', owner='John'))
        expected_data = {'age': 5, 'id': 1, 'name': 'Johnny', 'owner': 'John'}
        assert s_result.data == expected_data

    def test_abstraction(self):
        """Test that EntitySerializer class itself cannot be initialized"""

        with pytest.raises(ConfigurationError):
            EntitySerializer()

    def test_include_fields(self):
        """ Test the include fields option of the serializer"""

        class DogSerializer2(EntitySerializer):
            """ Serializer for the Dog Entity """
            class Meta:
                entity = Dog
                fields = ('id', 'age')

        s = DogSerializer2()
        assert s is not None

        # Check that the entity gets serialized correctly
        s_result = s.dump(Dog(id=1, name='Johnny', owner='John'))
        expected_data = {'age': 5, 'id': 1}
        assert s_result.data == expected_data

    def test_exclude_fields(self):
        """ Test the exclude fields option of the serializer"""

        class DogSerializer2(EntitySerializer):
            """ Serializer for the Dog Entity """
            class Meta:
                entity = Dog
                exclude = ('id', 'age')

        s = DogSerializer2()
        assert s is not None

        # Check that the entity gets serialized correctly
        s_result = s.dump(Dog(id=1, name='Johnny', owner='John'))
        expected_data = {'name': 'Johnny', 'owner': 'John'}
        assert s_result.data == expected_data

    def test_method_fields(self):
        """ Test the method field type of the serializer"""

        class DogSerializer2(EntitySerializer):
            """ Serializer for the Dog Entity """
            old = ma.fields.Method('get_old')

            def get_old(self, obj):
                """ Check if the dog is old or young """
                if obj.age > 5:
                    return True
                else:
                    return False

            class Meta:
                entity = Dog

        s = DogSerializer2()
        assert s is not None

        # Check that the entity gets serialized correctly
        s_result = s.dump(Dog(id=1, name='Johnny', owner='John'))
        expected_data = {
            'name': 'Johnny',
            'owner': 'John',
            'age': 5,
            'id': 1,
            'old': False
        }
        assert s_result.data == expected_data



"""Tests for Protean Flask's main method"""

from click.testing import CliRunner

from protean_flask.cli import main

from .support.sample_app import app


def test_main():
    """Test Main method"""
    runner = CliRunner()
    result = runner.invoke(main, [])

    assert 'Utility commands for the Protean-Flask package' in result.output
    assert result.exit_code == 0


def test_protean_context():
    """ Test that the protean context value gets set"""
    client = app.test_client()
    rv = client.get('/current-context')
    assert rv.status_code == 200
    assert rv.json == {
        'host_url': 'http://localhost/',
        'remote_addr': '127.0.0.1',
        'tenant_id': 'localhost',
        'url': 'http://localhost/current-context',
        'user_agent': 'werkzeug/0.14.1',
        'user_agent_hash': '11372f4fe6d226c4bb7fbd038c9c11d7d5ee400b847c9'
                           'fb4f37fe7642b25e182'
    }



"""
Default settings. Override these with settings in the module pointed to
by the PROTEAN_CONFIG environment variable.
"""

####################
# CORE             #
####################

DEBUG = False

# A secret key for this particular Protean installation. Used in secret-key
# hashing algorithms.
SECRET_KEY = 'abcdefghijklmn'

# Flag indicates that we are testing
TESTING = True

# Define the repositories
REPOSITORIES = {
    'default': {
        'PROVIDER': 'protean.impl.repository.dict_repo'
    }
}



"""Module to test View functionality and features"""

import json

import pytest
from protean.core.exceptions import ObjectNotFoundError
from protean.core.repository import repo
from tests.support.sample_app import app


class TestGenericAPIResource:
    """Class to test Generic API Resource functionality and methods"""

    @classmethod
    def setup_class(cls):
        """ Setup for this test case"""

        # Create the test client
        cls.client = app.test_client()

    def test_show(self):
        """ Test retrieving an entity using ShowAPIResource"""

        # Create a dog object
        repo.DogSchema.create(id=5, name='Johnny', owner='John')

        # Fetch this dog by ID
        rv = self.client.get('/dogs/5')
        assert rv.status_code == 200

        expected_resp = {
            'dog': {'age': 5, 'id': 5, 'name': 'Johnny', 'owner': 'John'}
        }
        assert rv.json == expected_resp

        # Test search by invalid id
        rv = self.client.get('/dogs/6')
        assert rv.status_code == 404

        # Delete the dog now
        repo.DogSchema.delete(5)

    def test_list(self):
        """ Test listing an entity using ListAPIResource """
        # Create a dog objects
        repo.DogSchema.create(id=1, name='Johnny', owner='John')
        repo.DogSchema.create(id=2, name='Mary', owner='John', age=3)
        repo.DogSchema.create(id=3, name='Grady', owner='Jane', age=8)
        repo.DogSchema.create(id=4, name='Brawny', owner='John', age=2)

        # Get the list of dogs
        rv = self.client.get('/dogs?order_by[]=age')
        assert rv.status_code == 200
        assert rv.json['total'] == 4
        assert rv.json['dogs'][0] == {'age': 2, 'id': 4, 'name': 'Brawny', 'owner': 'John'}

        # Test with filters
        rv = self.client.get('/dogs?owner=Jane')
        assert rv.status_code == 200
        assert rv.json['total'] == 1

    def test_create(self):
        """ Test creating an entity using CreateAPIResource """

        # Create a dog object
        rv = self.client.post('/dogs',
                              data=json.dumps(
                                  dict(id=5, name='Johnny', owner='John')),
                              content_type='application/json')
        assert rv.status_code == 201

        expected_resp = {
            'dog': {'age': 5, 'id': 5, 'name': 'Johnny', 'owner': 'John'}
        }
        assert rv.json == expected_resp

        # Test value has been added to db
        dog = repo.DogSchema.get(5)
        assert dog is not None
        assert dog.id == 5

        # Delete the dog now
        repo.DogSchema.delete(5)

    def test_update(self):
        """ Test updating an entity using UpdateAPIResource """

        # Create a dog object
        repo.DogSchema.create(id=5, name='Johnny', owner='John')

        # Update the dog object
        rv = self.client.put('/dogs/5',
                             data=json.dumps(dict(age=3)),
                             content_type='application/json')
        assert rv.status_code == 200

        expected_resp = {
            'dog': {'age': 3, 'id': 5, 'name': 'Johnny', 'owner': 'John'}
        }
        assert rv.json == expected_resp

        # Test value has been updated in the db
        dog = repo.DogSchema.get(5)
        assert dog is not None
        assert dog.age == 3

        # Delete the dog now
        repo.DogSchema.delete(5)

    def test_delete(self):
        """ Test deleting an entity using DeleteAPIResource """

        # Create a dog object
        repo.DogSchema.create(id=5, name='Johnny', owner='John')

        # Delete the dog object
        rv = self.client.delete('/dogs/5')
        assert rv.status_code == 204
        # print(dir(rv))
        assert rv.data == b''

        # Test value has been updated in the db
        with pytest.raises(ObjectNotFoundError):
            repo.DogSchema.get(5)


def test_flask_view():
    """ Test that non Protean views work as before """
    # Create the test client
    client = app.test_client()

    rv = client.get('/flask-view')
    assert rv.status_code == 200
    assert rv.data == b'View Response'

    rv = client.get('/flask-view/abc')
    assert rv.status_code == 404


def test_exception():
    """ Test handling of exceptions by the app """



"""Module to test request data parsing functionality"""
import json
from io import BytesIO

from flask import jsonify
from flask import request
from tests.support.sample_app import app
from werkzeug.datastructures import FileStorage

from protean_flask.core.views import APIResource


class DummyView(APIResource):
    """ Dummy view for testing the requests """
    methods = ['GET', 'POST', 'PUT']

    def get(self):
        """ Return the query params as the payload"""
        return jsonify(request.payload)

    def post(self):
        """ Return the form data/json as the payload"""
        for key, val in request.payload.items():
            if isinstance(val, FileStorage):
                request.payload[key] = val.filename
        return jsonify(request.payload)

    def put(self):
        """ Return the form data/json as the payload"""
        return jsonify(request.payload)


app.add_url_rule('/dummy', view_func=DummyView.as_view('dummy'))


class TestRequestParsing:
    """Tests for parsing of requests"""

    @classmethod
    def setup_class(cls):
        """ Setup for this test case"""

        # Create the test client
        cls.client = app.test_client()

        cls.payload = {'name': 'Harry', 'tags': [1, 2]}

    def test_form_type(self):
        """ Test parsing data when there is no content type """
        rv = self.client.post('/dummy', data=self.payload,
                              content_type=None)
        assert rv.status_code == 200

        payload = {'name': 'Harry', 'tags': ['1', '2']}
        assert rv.json == payload

    def test_default_type(self):
        rv = self.client.post('/dummy', data=json.dumps(self.payload),
                              content_type=None)
        assert rv.status_code == 200
        assert rv.json == self.payload

    def test_json_type(self):
        rv = self.client.post('/dummy', data=json.dumps(self.payload),
                              content_type='application/json')
        assert rv.status_code == 200
        assert rv.json == self.payload

    def test_query_params(self):
        rv = self.client.get('/dummy', query_string=self.payload)
        assert rv.status_code == 200

        payload = {'name': 'Harry', 'tags': ['1', '2']}
        assert rv.json == payload

    def test_multipart_type(self):
        """ Test parsing data when there is no content type """
        m_payload = {
            'file': (BytesIO(b'my file contents'), 'test file.txt')
        }
        m_payload.update(self.payload)
        rv = self.client.post('/dummy', data=m_payload,
                              content_type='multipart/form-data')
        assert rv.status_code == 200

        payload = {'name': 'Harry', 'tags': ['1', '2'],
                   'file': 'test file.txt'}
        assert rv.json == payload



"""This module exposes a generic Viewset class"""
from flask import request
from protean.core.usecase import CreateRequestObject
from protean.core.usecase import CreateUseCase
from protean.core.usecase import DeleteRequestObject
from protean.core.usecase import DeleteUseCase
from protean.core.usecase import ListRequestObject
from protean.core.usecase import ListUseCase
from protean.core.usecase import ShowRequestObject
from protean.core.usecase import ShowUseCase
from protean.core.usecase import UpdateRequestObject
from protean.core.usecase import UpdateUseCase

from protean_flask.core.views import GenericAPIResource


class GenericAPIResourceSet(GenericAPIResource):
    """This is a Generic View Set that has all the basic CRUD operations
    """
    show_usecase = ShowUseCase
    show_request_object = ShowRequestObject
    list_usecase = ListUseCase
    list_request_object = ListRequestObject
    create_usecase = CreateUseCase
    create_request_object = CreateRequestObject
    update_usecase = UpdateUseCase
    update_request_object = UpdateRequestObject
    delete_usecase = DeleteUseCase
    delete_request_object = DeleteRequestObject

    def get(self, identifier=None):
        """List the entities or Get by the identifier.
        """
        if identifier:
            payload = {'identifier': identifier}
            return self._process_request(
                self.show_usecase, self.show_request_object, payload=payload)
        else:
            return self._process_request(
                self.list_usecase, self.list_request_object,
                payload=request.payload, many=True)

    def post(self):
        """Create the entity.
        """
        return self._process_request(
            self.create_usecase, self.create_request_object,
            payload=request.payload)

    def put(self, identifier):
        """Update the entity.
         Expected Parameters:
             identifier = <string>, identifies the entity
        """
        payload = {
            'identifier': identifier,
            'data': request.payload
        }
        return self._process_request(
            self.update_usecase, self.update_request_object, payload=payload)

    def delete(self, identifier):
        """Delete the entity.
         Expected Parameters:
             identifier = <string>, identifies the entity
        """
        payload = {'identifier': identifier}
        return self._process_request(
            self.delete_usecase, self.delete_request_object, payload=payload)



"""Module that defines entry point to the Protean Flask Application"""
import hashlib

from flask import Blueprint
from flask import Request
from flask import current_app
from flask import request
from protean.conf import active_config
from protean.context import context
from protean.core.exceptions import UsecaseExecutionError
from protean.core.repository import repo
from protean.utils.importlib import perform_import

from ..utils import derive_tenant
from .views import APIResource


class ProteanRequest(Request):
    """ Custom request object to store protean specific code"""
    payload = None


class Protean(object):
    """
    The main entry point for the application.
    You need to initialize it with a Flask Application.

    >>> app = Flask(__name__)
    >>> api = Protean(app)

    Alternatively, you can use :meth:`init_app` to set the Flask application
    after it has been constructed.

    :param app_or_bp: the Flask application or blueprint object.

    """

    def __init__(self, app_or_bp=None):
        self.app = None
        self.blueprint = None
        self.exception_handler = None
        self.viewsets = []

        if app_or_bp is not None:
            self.app = app_or_bp
            if isinstance(app_or_bp, Blueprint):
                self.blueprint = app_or_bp
            else:
                self.init_app(app_or_bp)

    def init_app(self, app):
        """Perform initialization actions with the given :class:`flask.Flask`
        object.

        :param app: The flask application object
        :type app: flask.Flask
        """
        # Update the request class for the app
        app.request_class = ProteanRequest

        # Manage the protean information before/after request
        app.before_request(self._load_protean)
        app.after_request(self._cleanup_protean)

        # Register error handlers for the app
        app.register_error_handler(
            UsecaseExecutionError, self._handle_exception)
        self.exception_handler = perform_import(active_config.EXCEPTION_HANDLER)

        # Update the current configuration
        app.config.from_object(active_config)

    def register_viewset(self, view, endpoint, url, pk_name='identifier',
                         pk_type='string', additional_routes=None):
        """Register a Viewset

        Additional routes (apart from the standard five) can be specified via
            `additional_routes` argument. Note that the route names have to be
            the same as method names
        """
        # add the custom routes to the app
        if additional_routes is None:
            additional_routes = list()
        for route in additional_routes:
            route_name = endpoint
            if isinstance(route, (list, tuple)):
                route, route_name = route
            self.app.add_url_rule(
                '{}{}'.format(url, route), view_func=view.as_view(route_name))

        # Standard routes
        self.app.add_url_rule(
            url, view_func=view.as_view(f'list_{endpoint}'),
            methods=['GET', ])
        self.app.add_url_rule(
            url, view_func=view.as_view(f'create_{endpoint}'),
            methods=['POST', ])

        # Make sure that the url ends with a
        url = f'{url}/' if not url.endswith('/') else url
        self.app.add_url_rule('%s<%s:%s>' % (url, pk_type, pk_name),
                              view_func=view.as_view(f'show_{endpoint}'),
                              methods=['GET'])
        self.app.add_url_rule('%s<%s:%s>' % (url, pk_type, pk_name),
                              view_func=view.as_view(f'update_{endpoint}'),
                              methods=['PUT'])
        self.app.add_url_rule('%s<%s:%s>' % (url, pk_type, pk_name),
                              view_func=view.as_view(f'delete_{endpoint}'),
                              methods=['DELETE'])

    @staticmethod
    def _load_protean():
        """ Load the protean context with details from the request"""

        user_agent = request.headers.get('User-Agent', '')
        hashed_user_agent = hashlib.sha256(user_agent.encode())

        details = {
            'host_url': request.host_url,
            'url': request.url,
            'tenant_id': derive_tenant(request.url),
            'user_agent': user_agent,
            'user_agent_hash': hashed_user_agent.hexdigest(),
            'remote_addr': request.remote_addr
        }
        context.set_context(details)

    @staticmethod
    def _cleanup_protean(response):
        """ Cleanup the context and connections on end of request"""
        context.cleanup()
        repo.close_connections()
        return response

    def _handle_exception(self, e):
        """ Handle Protean exceptions and return appropriate response """

        # Get the renderer from the view class
        renderer = perform_import(active_config.DEFAULT_RENDERER)
        if request.url_rule:
            view_func = current_app.view_functions[request.url_rule.endpoint]
            view_class = view_func.view_class
            if issubclass(view_class, APIResource):
                renderer = getattr(view_class, 'renderer', renderer)

        # If user has defined an exception handler then call that
        if self.exception_handler:
            code, data, headers = self.exception_handler(e)

        else:
            code = e.value[0].value
            data = e.value[1]
            headers = {}

        # Build the response and return it
        response = renderer(data, code, headers)
        return response



"""Module to test Viewset functionality and features"""
import json

from protean.core.repository import repo
from tests.support.sample_app import app


class TestGenericAPIResourceSet:
    """Class to test GenericAPIResourceSet functionality and methods"""

    @classmethod
    def setup_class(cls):
        """ Setup for this test case"""

        # Create the test client
        cls.client = app.test_client()

    def test_set_show(self):
        """ Test retrieving an entity using the resource set"""
        # Create a human object
        repo.HumanSchema.create(id=1, name='John')

        # Fetch this human by ID
        rv = self.client.get('/humans/1')
        assert rv.status_code == 200
        expected_resp = {
            'human': {'contact': None, 'id': 1, 'name': 'John'}
        }
        assert rv.json == expected_resp

        # Delete the human now
        repo.HumanSchema.delete(1)

    def test_set_list(self):
        """ Test listing an entity using the resource set"""
        # Create Human objects
        repo.HumanSchema.create(id=2, name='Jane')
        repo.HumanSchema.create(id=3, name='Mary')

        # Get the list of humans
        rv = self.client.get('/humans?order_by[]=id')
        assert rv.status_code == 200
        assert rv.json['total'] == 2
        assert rv.json['humans'][0] == {'id': 2, 'name': 'Jane', 'contact': None}

    def test_set_create(self):
        """ Test creating an entity using the resource set """

        # Create a human object
        rv = self.client.post('/humans',
                              data=json.dumps(
                                  dict(id=1, name='John')),
                              content_type='application/json')
        assert rv.status_code == 201

        expected_resp = {
            'human': {'contact': None, 'id': 1, 'name': 'John'}
        }
        assert rv.json == expected_resp

        # Delete the human now
        repo.HumanSchema.delete(1)

    def test_set_update(self):
        """ Test updating an entity using the resource set """

        # Create a human object
        repo.HumanSchema.create(id=1, name='John')

        # Update the human object
        rv = self.client.put('/humans/1',
                             data=json.dumps(dict(contact='9000900090')),
                             content_type='application/json')
        assert rv.status_code == 200

        expected_resp = {
            'human': {'contact': '9000900090', 'id': 1, 'name': 'John'}
        }
        assert rv.json == expected_resp

        # Delete the human now
        repo.HumanSchema.delete(1)

    def test_set_delete(self):
        """ Test deleting an entity using the resource set """

        # Create a human object
        repo.HumanSchema.create(id=1, name='John')

        # Delete the dog object
        rv = self.client.delete('/humans/1')
        assert rv.status_code == 204
        assert rv.data == b''

    def test_custom_route(self):
        """ Test custom routes using the resource set """

        # Create a human object
        repo.HumanSchema.create(id=1, name='John')

        # Get the custom route
        rv = self.client.get('/humans/1/my_dogs')
        assert rv.status_code == 200
        assert rv.json['total'] == 3
        assert rv.json['dogs'][0] == {'age': 2, 'id': 4, 'name': 'Brawny', 'owner': 'John'}
